/* schur_id3b.c: Generic Mex-C program for computing the Schur algorithm
 * with Gaussian Process.
 *
 * Syntax used:
 *			[PR,SD] = schur_id3b(G,b)
 *
 * Inputs:
 * 		G	: Schur Generator matrix
 *		b	: Any column vector
 *
 * Output:
 * 		PR	: Returns the prediction
 *		SD	: Returns the standard deviation
 *
 * Generates the prediction and standard deviations of the
 * Gaussian process using Generalised Schur algorithm, where the prediction
 * and standard deviation are defined as dQ'*inv(T)*b and
 * sqrt{diag {AQ-dQ'*inv(T)*dQ} } respectively. The output of the standard
 * deviation is only diag{dQ'*inv(T)*dQ}
 *
 *      [-T   dQ -b]
 *  R = [dQ'  AQ  0]
 *      [-b'  0   0]
 *
 * where T is positive-definite strongly Toeplitz-like Block and dQ is the
 * cross-covariance matrix between the explanatory variables and the
 * predicted explanatory variables. AQ is the covariance matrix between the
 * predicted explantory variables and itself. Note that no hyperbolic
 * rotation is implemented.
 *
 * For use with Matlab version 7.0 and above. Other versions may be possible,
 * but not tested.
 *
 * Instruction: Run <mex xxx.c> at Matlab command prompt to compile
 * the .c code into .dll (Windows) or .meglx (Linux/UNIX).
 *
 *
 * (C) Gaussian Process Schur Toolbox 1.0
 * (C) Copyright 2005-2007, Keith Neo <keith.neo@nuim.ie>
 * 							http://www.hamilton.ie */

#include "mex.h"

void pred_schur(unsigned long n,unsigned long rk,unsigned long n3,double *G,double *b,double *pr,double *sd)
{
	unsigned long k,j,h,nr,np,nk,nk1,nz3,rk2;
	double tmp,d;
	double *Gen,*g,*l;

	Gen=(double *)malloc(n*rk*sizeof(double));
	for (k=0;k<rk;k++)
	{
		nr=k*n;
		for (j=0;j<n;j++)
		{
			np=nr+j;
			Gen[np]=G[np];
		}
	}

	/*prediction size*/
	nr=n-n3-1;
	rk2=rk/2;
	l=(double *)malloc(n*sizeof(double));
	g=(double *)malloc(rk*sizeof(double));
	for (k=0;k<n3;k++)
	{
		nk=n-k;
		nk1=nk-1;
		for (j=0;j<rk;j++) g[j]=Gen[j*nk];

		for (j=0;j<nk;j++)
		{
			tmp=0.0;
			for (h=0;h<rk;h++)
			{
				d=Gen[h*nk+j]*g[h];
				if (h<rk2) tmp=tmp+d;
				else tmp=tmp-d;
			}
			l[j]=tmp;
		}
		d=-1.0/l[0];

		nz3=n3-k;
		for (j=0;j<nr;j++)
		{
			np=nz3+j;
			sd[j]=sd[j]-l[np]*l[np]*d;
			pr[j]=pr[j]+l[np]*l[nk1]*d;
		}

		for (j=(nk-1);j>0;j--)
		{
			if ((j==nz3) || (j==nk1)) l[j]=-l[j];
			else l[j]=l[j-1]-l[j];
		}

		for (j=0;j<rk;j++)
		{
			tmp=g[j]*d;
			nz3=j*nk;
			np=nz3-j-1;
			for (h=1;h<nk;h++) Gen[np+h]=Gen[nz3+h]-l[h]*tmp;
		}
	}
	free(g);
	free(l);
	free(Gen);
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
	double *PR;
	double *SD;
	double *G;
	double *B;
	unsigned long NG;
	unsigned long R;
	unsigned long N;

  if(nrhs!=2) {
    mexErrMsgTxt("Two inputs required.");
  } else if(nlhs>2) {
    mexErrMsgTxt("Too many  output arguments");
  }

  NG = mxGetM(prhs[0]);
  R  = mxGetN(prhs[0]);
  N  = mxGetM(prhs[1]);

  plhs[0] = mxCreateDoubleMatrix(NG-N-1,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(NG-N-1,1, mxREAL);

  G  = mxGetPr(prhs[0]);
  B  = mxGetPr(prhs[1]);
  PR = mxGetPr(plhs[0]);
  SD = mxGetPr(plhs[1]);

  pred_schur(NG,R,N,G,B,PR,SD);
  return;
}
