/* schur_id3b_r.c: Generic Mex-C program for computing the Schur algorithm
 * with Gaussian Process.
 *
 * Syntax used:
 *			[PR,SD] = schur_id3b_r(G,b)
 *
 * Inputs:
 * 		G	: Schur Generator matrix
 *		b	: Any column vector
 *
 * Output:
 * 		PR	: Returns the prediction
 *		SD	: Returns the standard deviation
 *
 * Generates the prediction and standard deviations of the
 * Gaussian process using Generalised Schur algorithm, where the prediction
 * and standard deviation are defined as dQ'*inv(T)*b and
 * sqrt{diag {AQ-dQ'*inv(T)*dQ} } respectively. The output of the standard
 * deviation is only diag{dQ'*inv(T)*dQ}
 *
 *      [-T   dQ -b]
 *  R = [dQ'  AQ  0]
 *      [-b'  0   0]
 *
 * where T is positive-definite strongly Toeplitz-like Block and dQ is the
 * cross-covariance matrix between the explanatory variables and the
 * predicted explanatory variables. AQ is the covariance matrix between the
 * predicted explantory variables and itself. Note, SVD-type hyperbolic
 * rotation is implemented for any rank, i.e. rank(disp.R)=4.
 *
 * For use with Matlab version 7.0 and above. Other versions may be possible,
 * but not tested.
 *
 * Instruction: Run <mex xxx.c> at Matlab command prompt to compile
 * the .c code into .dll (Windows) or .meglx (Linux/UNIX).
 *
 *
 * (C) Gaussian Process Schur Toolbox 1.0
 * (C) Copyright 2005-2007, Keith Neo <keith.neo@nuim.ie>
 * 							http://www.hamilton.ie */

#include "mex.h"
#define abs(x) (((x)>0)?(x):-(x))

void pred_schur(unsigned long,unsigned long,unsigned long,double *,double *,double *,double *);
void hyper_rotate(unsigned long,unsigned long,double *);

void pred_schur(unsigned long n,unsigned long rk,unsigned long n3,double *G,double *b,double *pr,double *sd)
{
	unsigned long k,j,h,nr,np,nk,nk1,nz3,rk2;
	double tmp,d;
	double *Gen,*g,*l;

	Gen=(double *)malloc(n*rk*sizeof(double));
	for (k=0;k<rk;k++)
	{
		nr=k*n;
		for (j=0;j<n;j++)
		{
			np=nr+j;
			Gen[np]=G[np];
		}
	}

	nr=n-n3-1;
	rk2=rk/2;
	l=(double *)malloc(n*sizeof(double));
	g=(double *)malloc(rk*sizeof(double));
	for (k=0;k<n3;k++)
	{
		nk=n-k;
		nk1=nk-1;
		hyper_rotate(rk,nk,Gen);
		for (j=0;j<rk;j++) g[j]=Gen[j*nk];

		for (j=0;j<nk;j++)
		{
			tmp=0.0;
			for (h=0;h<rk;h++)
			{
				d=Gen[h*nk+j]*g[h];
				if (h<rk2) tmp=tmp+d;
				else tmp=tmp-d;
			}
			l[j]=tmp;
		}
		d=-1.0/l[0];

		nz3=n3-k;
		for (j=0;j<nr;j++)
		{
			np=nz3+j;
			sd[j]=sd[j]-l[np]*l[np]*d;
			pr[j]=pr[j]+l[np]*l[nk1]*d;
		}

		for (j=(nk-1);j>0;j--)
		{
			if ((j==nz3) || (j==nk1)) l[j]=-l[j];
			else l[j]=l[j-1]-l[j];
		}

		for (j=0;j<rk;j++)
		{
			tmp=g[j]*d;
			nz3=j*nk;
			np=nz3-j-1;
			for (h=1;h<nk;h++) Gen[np+h]=Gen[nz3+h]-l[h]*tmp;
		}
	}
	free(g);
	free(l);
	free(Gen);
	return;
}

void hyper_rotate(unsigned long rk,unsigned long nk,double *G)
{
	unsigned long k,j,rk2,np;
	double ma=0.0;
	double mb=0.0;
	double tmp,tmp1,p;
	double *xa,*xb,*g,*r;

	rk2=rk/2;
	np=rk2-1;
	xa=(double *)malloc(rk2*sizeof(double));
	xb=(double *)malloc(rk2*sizeof(double));
	for (k=0;k<rk2;k++)
	{
		xa[k]=G[k*nk];
		xb[k]=G[(rk2+k)*nk];
		if (k>0) ma=ma+xa[k]*xa[k];
		if (k<np) mb=mb+xb[k]*xb[k];
	}
	tmp=ma+xa[0]*xa[0];
	tmp1=mb+xb[np]*xb[np];
	xa[0]=xa[0]+sqrt(tmp);
	xb[np]=xb[np]+sqrt(tmp1);
	p=tmp-tmp1;

	ma=ma+xa[0]*xa[0];
	mb=mb+xb[np]*xb[np];

	g=(double *)malloc(rk2*sizeof(double));
	r=(double *)malloc(nk*sizeof(double));
	if (ma>1.0e-307)
	{
		tmp1=2.0/ma;
		for (k=0;k<rk2;k++) g[k]=xa[k]*tmp1;
		for (k=0;k<nk;k++)
		{
			tmp=0.0;
			for (j=0;j<rk2;j++) tmp=tmp+G[j*nk+k]*xa[j];
			r[k]=tmp;
		}

		for (k=0;k<rk2;k++)
		{
			np=k*nk;
			for(j=0;j<nk;j++) G[np+j]=G[np+j]-r[j]*g[k];
		}
	}

	if (mb>1.0e-307)
	{
		tmp1=2.0/mb;
		for (k=0;k<rk2;k++) g[k]=xb[k]*tmp1;
		for (k=0;k<nk;k++)
		{
			tmp=0.0;
			for (j=0;j<rk2;j++) tmp=tmp+G[(rk2+j)*nk+k]*xb[j];
			r[k]=tmp;
		}

		for (k=0;k<rk2;k++)
		{
			np=(rk2+k)*nk;
			for (j=0;j<nk;j++) G[np+j]=G[np+j]-r[j]*g[k];
		}
	}
	free(xa);
	free(xb);
	free(g);
	free(r);

	np=(rk-1)*nk;
	ma=abs(G[0]+G[np]);
	mb=abs(G[0]-G[np]);
	if ((ma!=0.0) || (mb!=0.0))
	{
		tmp=0.5*sqrt(ma/mb);
		tmp1=0.5*sqrt(mb/ma);
		for (k=0;k<nk;k++)
		{
			ma=tmp*(G[k]-G[np+k]);
			mb=tmp1*(G[k]+G[np+k]);
			G[k]=mb+ma;
			G[np+k]=mb-ma;
		}
	}
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
	double *PR;
	double *SD;
	double *G;
	double *B;
	unsigned long NG;
	unsigned long R;
	unsigned long N;

  if(nrhs!=2) {
    mexErrMsgTxt("Two inputs required.");
  } else if(nlhs>2) {
    mexErrMsgTxt("Too many  output arguments");
  }

  NG = mxGetM(prhs[0]);
  R  = mxGetN(prhs[0]);
  N  = mxGetM(prhs[1]);

  plhs[0] = mxCreateDoubleMatrix(NG-N-1,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(NG-N-1,1, mxREAL);

  G  = mxGetPr(prhs[0]);
  B  = mxGetPr(prhs[1]);
  PR = mxGetPr(plhs[0]);
  SD = mxGetPr(plhs[1]);

  pred_schur(NG,R,N,G,B,PR,SD);
  return;
}
