/* schur_invbfv3.c: Generic Mex-C program for computing the Schur algorithm
 * with Gaussian Process.
 *
 * Syntax used:
 *			[LOGDET,YQY,TR,YQQY,INVB] = schur_invbfv3(G,b)
 *
 * Inputs:
 * 		G	: Schur Generator matrix
 *		b	: Any column vector
 *
 * Output:
 * 		LOGDET	: Returns the value of the log-determinant
 *		YQY		: Returns the value of b'*inv(T)*b
 *		TR		: Returns the value of the trace operation of inv(T)
 *		YQQY	: Returns the value of b'*inv(T)*inv(T)*b
 *		INVB	: Returns the column vector of inv(T)*b
 *
 * Generates the Generalise Schur' complement of
 * the Hermitian matrix, R, such that
 *
 * R=[-T  I]
 *   [ I  0]
 *
 * where T is the positive-definite Toeplitz-like Block matrix. Note
 * that no hyperbolic rotation is implemented.
 *
 * For use with Matlab version 7.0 and above. Other versions may be possible,
 * but not tested.
 *
 * Instruction: Run <mex xxx.c> at Matlab command prompt to compile
 * the .c code into .dll (Windows) or .meglx (Linux/UNIX).
 *
 *
 * (C) Gaussian Process Schur Toolbox 1.0
 * (C) Copyright 2005-2007, Keith Neo <keith.neo@nuim.ie>
 * 							http://www.hamilton.ie */

#include <math.h>
#include "mex.h"
#define abs(x) (((x)>0)?(x):-(x))

void trace_invfast3(unsigned long n,unsigned long rk,double *G,double *b,double *logdet,double *yQy,double *trq2,double *yQQy,double *invb)
{
	unsigned long k,j,h,n2,np,nk,nz,nx,rk2;
	double tmp,tmp2,d;
	double *Gen,*g,*l;

	n2=n/2;
	nx=n-2;
	Gen=(double *)malloc(n*rk*sizeof(double));
	for (k=0;k<rk;k++)
	{
		nz=k*n;
		for (j=0;j<n;j++)
		{
			np=nz+j;
			if (j<=nx) Gen[np]=G[np];
			else Gen[np]=0.0;
		}
	}

	nx=nx+2;
	rk2=rk/2;
	logdet[0]=0.0;
	l=(double *)malloc(n*sizeof(double));
	g=(double *)malloc(rk*sizeof(double));
	for (k=0;k<n2;k++)
	{
		nk=n-k;
		for (j=0;j<rk;j++) g[j]=Gen[j*nk];

		for (j=0;j<nk;j++)
		{
			tmp=0.0;
			for (h=0;h<rk;h++)
			{
				d=Gen[h*nk+j]*g[h];
				if (h<rk2) tmp=tmp+d;
				else tmp=tmp-d;
			}
			l[j]=tmp;
		}
		logdet[0]=logdet[0]+log(abs(l[0]));
		d=-1.0/l[0];
		tmp=b[k];
		tmp2=1.0;
		h=nk-n2;
		for (j=0;j<k;j++)
		{
			np=h+j;
			tmp=tmp+l[np]*b[j];
			tmp2=tmp2+l[np]*l[np];
		}
		trq2[0]=trq2[0]+tmp2*d;
		invb[k]=tmp*d;
		for (j=0;j<k;j++) invb[j]=invb[j]+l[h+j]*invb[k];

		np=h-n2;
		for (j=(nk-1);j>0;j--)
		{
			if ((j==h) || (j==np)) l[j]=-l[j];
			else l[j]=l[j-1]-l[j];
		}

		for (j=0;j<rk;j++)
		{
			tmp=g[j]*d;
			nz=j*nk;
			np=nz-j-1;
			for (h=1;h<nk;h++)
			{
				if (h<nx) Gen[np+h]=Gen[nz+h]-l[h]*tmp;
				else Gen[np+h]=0.0;
			}
		}
	}
	free(g);
	free(l);
	free(Gen);
	yQy[0]=0.0;
	yQQy[0]=0.0;
	for (k=0;k<n2;k++)
	{
		yQy[0]=yQy[0]+invb[k]*b[k];
		yQQy[0]=yQQy[0]+invb[k]*invb[k];
	}
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
	double *LDT;
	double *BQB;
	double *TRQ;
	double *QB2;
	double *INB;
	double *G;
	double *B;
	unsigned long N;
	unsigned long R;

  if(nrhs!=2) {
    mexErrMsgTxt("Two inputs required.");
  } else if(nlhs>5) {
    mexErrMsgTxt("Too many  output arguments");
  }

  N = mxGetM(prhs[0]);
  R = mxGetN(prhs[0]);

  plhs[0] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[2] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[3] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[4] = mxCreateDoubleMatrix(N/2,1, mxREAL);

  G   = mxGetPr(prhs[0]);
  B   = mxGetPr(prhs[1]);
  LDT = mxGetPr(plhs[0]);
  BQB = mxGetPr(plhs[1]);
  TRQ = mxGetPr(plhs[2]);
  QB2 = mxGetPr(plhs[3]);
  INB = mxGetPr(plhs[4]);

  trace_invfast3(N,R,G,B,LDT,BQB,TRQ,QB2,INB);
  return;
}
