/* schur_invbtrq3_hessian.c: Generic Mex-C program for computing the Schur algorithm
 * with Gaussian Process.
 *
 * Syntax used:
 *			[TRDQ,DQB]=schur_invbtrq3_hessian(G,b)
 *
 * Inputs:
 * 		G	: Schur Generator matrix
 *		b	: Any column vector
 *
 * Output:
 * 		TRDQ	: Returns the trace value of {inv(T)*[dQdx]}
 *		DQB		: Returns the vector of { [dQdx]*inv(T)*b }
 *
 * Generates the trace of {[dQdx]*inv(T)} using the Schur
 * complement of the Hermitian matrix, R, as well as returns vector
 * result of {[dQdx]*inv(T)*b}, such that
 *
 *     [-T     dQdx I]
 * R = [ dQdx  0    0]
 *     [ I     0    0]
 *
 * where T and dQdx are positive-definite strongly Toeplitz-like Block
 * matrix, except that only the diagonal of dQdx are zeros. Note that no
 * hyperbolic rotation is implemented if T is Toeplitz, i.e. rank(disp.R)=4
 *
 * For use with Matlab version 7.0 and above. Other versions may be possible,
 * but not tested.
 *
 * Instruction: Run <mex xxx.c> at Matlab command prompt to compile
 * the .c code into .dll (Windows) or .meglx (Linux/UNIX).
 *
 *
 * (C) Gaussian Process Schur Toolbox 1.0
 * (C) Copyright 2005-2007, Keith Neo <keith.neo@nuim.ie>
 * 							http://www.hamilton.ie */

#include "mex.h"

void trace_invschur2(unsigned long n,unsigned long rk,double *G,double *b,double *trq2,double *dQb)
{
	unsigned long k,j,h,n3,np,nk,nz,nx,rk2;
	double tmp,tmp1,d;
	double *Gen,*g,*l;

	n3=n/3;
	nx=n-3;
	Gen=(double *)malloc(n*rk*sizeof(double));
	for (k=0;k<rk;k++)
	{
		nz=k*n;
		for (j=0;j<n;j++)
		{
			np=nz+j;
			if (j<=nx) Gen[np]=G[np];
			else Gen[np]=0.0;
		}
	}

	nx=nx+2;
	trq2[0]=0.0;
	rk2=rk/2;
	l=(double *)malloc(n*sizeof(double));
	g=(double *)malloc(rk*sizeof(double));
	for (k=0;k<n3;k++)
	{
		nk=n-k;
		for (j=0;j<rk;j++) g[j]=Gen[j*nk];

		for (j=0;j<nk;j++)
		{
			tmp=0.0;
			for (h=0;h<rk;h++)
			{
				d=Gen[h*nk+j]*g[h];
				if (h<rk2) tmp=tmp+d;
				else tmp=tmp-d;
			}
			l[j]=tmp;
		}
		d=-1.0/l[0];

		tmp=b[k];
		tmp1=l[n3];
		h=nk-n3;
		for (j=0;j<k;j++)
		{
			np=h+j;
			tmp=tmp+l[np]*b[j];
			tmp1=tmp1+l[np-n3]*l[np];
		}
		trq2[0]=trq2[0]+tmp1*d;
		tmp1=tmp*d;
		h=n3-k;
		for (j=0;j<n3;j++) dQb[j]=dQb[j]+l[h+j]*tmp1;

		h=nk-n3;
		np=h-n3;
		for (j=(nk-1);j>0;j--)
		{
			if ((j==h) || (j==np)) l[j]=-l[j];
			else l[j]=l[j-1]-l[j];
		}

		for (j=0;j<rk;j++)
		{
			tmp=g[j]*d;
			nz=j*nk;
			np=nz-j-1;
			for (h=1;h<nk;h++)
			{
				if (h<nx) Gen[np+h]=Gen[nz+h]-l[h]*tmp;
				else Gen[np+h]=0.0;
			}
		}
	}
	free(g);
	free(l);
	free(Gen);
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
	double *TRQ;
	double *DQB;
	double *G;
	double *B;
	unsigned long N;
	unsigned long R;

  if(nrhs!=2) {
    mexErrMsgTxt("Two inputs required.");
  } else if(nlhs>2) {
    mexErrMsgTxt("Too many  output arguments");
  }

  N = mxGetM(prhs[0]);
  R = mxGetN(prhs[0]);

  plhs[0] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(N/3,1, mxREAL);

  G   = mxGetPr(prhs[0]);
  B   = mxGetPr(prhs[1]);
  TRQ = mxGetPr(plhs[0]);
  DQB = mxGetPr(plhs[1]);

  trace_invschur2(N,R,G,B,TRQ,DQB);
  return;
}
