/* schur_trq3dqdq_ar.c: Generic Mex-C program for computing the Schur algorithm
 * with Gaussian Process.
 *
 * Syntax used:
 *			[TRDQ,DQB,TRDQ3] = schur_trq3dqdq_ar(G,b,r)
 *
 * Inputs:
 * 		G	: Schur Generator matrix
 *		b	: Any column vector
 *		r	: Coefficient, scaling factor for stabilising pivoting matrix
 *
 * Output:
 *		TRDQ   = trace result from { [Q*inv(T)*Q*inv(T)] }
 *		DQB    = returns vector of { [Q*inv(T)*Q*inv(T)]*b }
 *		TRQ3   = trace result from { [inv(T)-r*inv(T)*Q*inv(T)] }
 *
 * Generates the trace of [r*Q*inv(T)*Q*inv(T)-Q*inv(T)]
 * using Schur complement of the Hermitian matrix, R, as well as returns
 * vector result of { [r*Q*inv(T)*Q*inv(T)-Q*inv(T)]*b }, such that
 *
 *     [ T-r*Q  T  -Q 0 ]
 * R = [   T    0  -Q I ]
 *     [  -Q   -Q   0 0 ]
 *     [   0    I   0 0 ]
 *
 * where T is a positive-definite strongly Toeplitz-like Block  matrix, and
 * Q is symmetrical, but may be positive or non-postive-definite
 * Toepliz-like matrix. In this case, we expect the diagonal of Q to be
 * zeros, since Q <==> dQdg in Gaussian Process. Hyperbolic rotation is implemented
 * in this algorithm.
 *
 * For use with Matlab version 7.0 and above. Other versions may be possible,
 * but not tested.
 *
 * Instruction: Run <mex xxx.c> at Matlab command prompt to compile
 * the .c code into .dll (Windows) or .meglx (Linux/UNIX).
 *
 *
 * (C) Gaussian Process Schur Toolbox 1.0
 * (C) Copyright 2005-2007, Keith Neo <keith.neo@nuim.ie>
 * 							http://www.hamilton.ie */

#include "mex.h"
#define abs(x) (((x)>0)?(x):-(x))

void trace_schur3(unsigned long,unsigned long,double *,double *,double *,double *,double *,double *);
void hyper_rotate(unsigned long,unsigned long,double *);

void trace_schur3(unsigned long n,unsigned long rk,double *G,double *b,double *coef,double *trq2,double *dQb,double *trq3)
{
	unsigned long k,j,h,n4,np,nk,nz,nx,nr,rk2;
	double tmp,tmp1,tmp2,d;
	double *Gen,*g,*l;

	n4=n/4;
	nx=n-n4;
	Gen=(double *)malloc(n*rk*sizeof(double));
	for (k=0;k<rk;k++)
	{
		nz=k*n;
		for (j=0;j<n;j++)
		{
			np=nz+j;
			if (j<=nx) Gen[np]=G[np];
			else Gen[np]=0.0;
		}
	}

	nx=nx+2;
	rk2=rk/2;
	trq2[0]=0.0;
	trq3[0]=0.0;
	l=(double *)malloc(n*sizeof(double));
	g=(double *)malloc(rk*sizeof(double));
	for (k=0;k<2*n4;k++)
	{
		nk=n-k;
		hyper_rotate(rk,nk,Gen);
		for (j=0;j<rk;j++) g[j]=Gen[j*nk];

		for (j=0;j<nk;j++)
		{
			tmp=0.0;
			for (h=0;h<rk;h++)
			{
				d=Gen[h*nk+j]*g[h];
				if (h<rk2) tmp=tmp+d;
				else tmp=tmp-d;
			}
			l[j]=tmp;
		}
		d=1.0/l[0];

		nz=nk-n4;
		nr=nz-n4;
		if (k>=n4)
		{
			h=k-n4;
			tmp=b[h];
			tmp1=l[n4];
			tmp2=1.0;
			for (j=0;j<h;j++)
			{
				np=j+nz;
				tmp1=tmp1+l[np-n4]*l[np];
				tmp2=tmp2+l[np]*l[np];
				tmp=tmp+l[np]*b[j];
			}
			tmp=tmp*d;
			trq2[0]=trq2[0]-tmp1*d;
			trq3[0]=trq3[0]-tmp2*d;
			for (j=0;j<n4;j++) dQb[j]=dQb[j]-l[j+nr]*tmp;
		}

		np=nr-n4;
		for (j=(nk-1);j>0;j--)
		{
			if ((j==nz) || (j==nr) || (j==np)) l[j]=-l[j];
			else l[j]=l[j-1]-l[j];
		}

		for (j=0;j<rk;j++)
		{
			tmp=g[j]*d;
			nz=j*nk;
			np=nz-j-1;
			for (h=1;h<nk;h++)
			{
				if (h<nx) Gen[np+h]=Gen[nz+h]+l[h]*tmp;
				else Gen[np+h]=0.0;
			}
		}
	}
	free(g);
	free(l);
	free(Gen);
	tmp=1/coef[0];
	for (k=0;k<n4;k++) dQb[k]=tmp*dQb[k];
	trq2[0]=trq2[0]*tmp;
	return;
}

void hyper_rotate(unsigned long rk,unsigned long nk,double *G)
{
	unsigned long k,j,rk2,np;
	double ma=0.0;
	double mb=0.0;
	double tmp,tmp1,p;
	double *xa,*xb,*g,*r;

	rk2=rk/2;
	np=rk2-1;
	xa=(double *)malloc(rk2*sizeof(double));
	xb=(double *)malloc(rk2*sizeof(double));
	for (k=0;k<rk2;k++)
	{
		xa[k]=G[k*nk];
		xb[k]=G[(rk2+k)*nk];
		if (k>0) ma=ma+xa[k]*xa[k];
		if (k<np) mb=mb+xb[k]*xb[k];
	}
	tmp=ma+xa[0]*xa[0];
	tmp1=mb+xb[np]*xb[np];
	xa[0]=xa[0]+sqrt(tmp);
	xb[np]=xb[np]+sqrt(tmp1);
	p=tmp-tmp1;

	ma=ma+xa[0]*xa[0];
	mb=mb+xb[np]*xb[np];

	g=(double *)malloc(rk2*sizeof(double));
	r=(double *)malloc(nk*sizeof(double));
	if (ma>1.0e-307)
	{
		tmp1=2.0/ma;
		for (k=0;k<rk2;k++) g[k]=xa[k]*tmp1;
		for (k=0;k<nk;k++)
		{
			tmp=0.0;
			for (j=0;j<rk2;j++) tmp=tmp+G[j*nk+k]*xa[j];
			r[k]=tmp;
		}

		for (k=0;k<rk2;k++)
		{
			np=k*nk;
			for(j=0;j<nk;j++) G[np+j]=G[np+j]-r[j]*g[k];
		}
	}

	if (mb>1.0e-307)
	{
		tmp1=2.0/mb;
		for (k=0;k<rk2;k++) g[k]=xb[k]*tmp1;
		for (k=0;k<nk;k++)
		{
			tmp=0.0;
			for (j=0;j<rk2;j++) tmp=tmp+G[(rk2+j)*nk+k]*xb[j];
			r[k]=tmp;
		}

		for (k=0;k<rk2;k++)
		{
			np=(rk2+k)*nk;
			for (j=0;j<nk;j++) G[np+j]=G[np+j]-r[j]*g[k];
		}
	}
	free(xa);
	free(xb);
	free(g);
	free(r);

	np=(rk-1)*nk;
	ma=abs(G[0]+G[np]);
	mb=abs(G[0]-G[np]);
	if ((ma!=0.0) || (mb!=0.0))
	{
		tmp=0.5*sqrt(ma/mb);
		tmp1=0.5*sqrt(mb/ma);
		for (k=0;k<nk;k++)
		{
			ma=tmp*(G[k]-G[np+k]);
			mb=tmp1*(G[k]+G[np+k]);
			G[k]=mb+ma;
			G[np+k]=mb-ma;
		}
	}
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
	double *TRQ;
	double *DQB;
	double *TRQ3;
	double *G;
	double *B;
	double *C;
	unsigned long N;
	unsigned long R;

  if(nrhs!=3) {
    mexErrMsgTxt("Three inputs required.");
  } else if(nlhs>3) {
    mexErrMsgTxt("Too many  output arguments");
  }

  N = mxGetM(prhs[0]);
  R = mxGetN(prhs[0]);

  plhs[0] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(N/4,1, mxREAL);
  plhs[2] = mxCreateDoubleMatrix(1,1, mxREAL);

  G   = mxGetPr(prhs[0]);
  B   = mxGetPr(prhs[1]);
  C   = mxGetPr(prhs[2]);
  TRQ = mxGetPr(plhs[0]);
  DQB = mxGetPr(plhs[1]);
  TRQ3= mxGetPr(plhs[2]);

  trace_schur3(N,R,G,B,C,TRQ,DQB,TRQ3);
  return;
}
