/* schur_trq3qq.c: Generic Mex-C program for computing the Schur algorithm
 * with Gaussian Process.
 *
 * Syntax used:
 *			[TRQ2,DQB] = schur_trace3qq(G,b)
 *
 * Inputs:
 * 		G	: Schur Generator matrix
 *		b	: Any column vector
 *
 * Output:
 *		TRQ2   = trace result from { inv(T*T) }
 *		DQB    = returns vector of { inv(T)*inv(T)*b }
 *
 * Generates the trace of inv(T*T) using the Schur
 * complement of the Hermitian matrix, R, as well as returns vector
 * result of {inv(T)*inv(T)*b}, such that
 *
 *     [ I  T  I ]
 * R = [ T  0  I ]
 *     [ 0  I  0 ]
 *
 * where T is a positive-definite strongly Toeplitz-like Block  matrix,
 * except that only the diagonal of dQdx are zeros. Note that no
 * hyperbolic rotation is implemented.
 *
 * For use with Matlab version 7.0 and above. Other versions may be possible,
 * but not tested.
 *
 * Instruction: Run <mex xxx.c> at Matlab command prompt to compile
 * the .c code into .dll (Windows) or .meglx (Linux/UNIX).
 *
 *
 * (C) Gaussian Process Schur Toolbox 1.0
 * (C) Copyright 2005-2007, Keith Neo <keith.neo@nuim.ie>
 * 							http://www.hamilton.ie */

#include "mex.h"

void trace_schur3qq(unsigned long n,unsigned long rk,double *G,double *b,double *trq2,double *dQb)
{
	unsigned long k,j,h,n3,np,nk,nz,nx,rk2;
	double tmp,tmp2,d;
	double *Gen,*g,*l;

	n3=n/3;
	nx=n-n3;
	Gen=(double *)malloc(n*rk*sizeof(double));
	for (k=0;k<rk;k++)
	{
		nz=k*n;
		for (j=0;j<n;j++)
		{
			np=nz+j;
			if (j<=nx) Gen[np]=G[np];
			else Gen[np]=0.0;
		}
	}

	nx=nx+2;
	trq2[0]=0.0;
	rk2=rk/2;
	l=(double *)malloc(n*sizeof(double));
	g=(double *)malloc(rk*sizeof(double));
	for (k=0;k<2*n3;k++)
	{
		nk=n-k;
		for (j=0;j<rk;j++) g[j]=Gen[j*nk];


		for (j=0;j<nk;j++)
		{
			tmp=0.0;
			for (h=0;h<rk;h++)
			{
				d=Gen[h*nk+j]*g[h];
				if (h<rk2) tmp=tmp+d;
				else tmp=tmp-d;
			}
			l[j]=tmp;
		}
		d=1.0/l[0];

		nz=nk-n3;
		if (k>=n3)
		{
			h=k-n3;
			tmp=b[h];
			tmp2=1.0;
			for (j=0;j<h;j++)
			{
				np=j+nz;
				tmp2=tmp2+l[np]*l[np];
				tmp=tmp+l[np]*b[j];
			}
			trq2[0]=trq2[0]-tmp2*d;
			dQb[h]=-tmp*d;
			for (j=0;j<h;j++) dQb[j]=dQb[j]+l[j+nz]*dQb[h];
		}

		np=nz-n3;
		for (j=(nk-1);j>0;j--)
		{
			if ((j==nz) || (j==np)) l[j]=-l[j];
			else l[j]=l[j-1]-l[j];
		}

		for (j=0;j<rk;j++)
		{
			tmp=g[j]*d;
			nz=j*nk;
			np=nz-j-1;
			for (h=1;h<nk;h++)
			{
				if (h<nx) Gen[np+h]=Gen[nz+h]+l[h]*tmp;
				else Gen[np+h]=0.0;
			}
		}
	}
	free(g);
	free(l);
	free(Gen);
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[] )
{
	double *TRQ;
	double *DQB;
	double *G;
	double *B;
	unsigned long N;
	unsigned long R;

  if(nrhs!=2) {
    mexErrMsgTxt("Two inputs required.");
  } else if(nlhs>2) {
    mexErrMsgTxt("Too many  output arguments");
  }

  N = mxGetM(prhs[0]);
  R = mxGetN(prhs[0]);

  plhs[0] = mxCreateDoubleMatrix(1,1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(N/3,1, mxREAL);

  G   = mxGetPr(prhs[0]);
  B   = mxGetPr(prhs[1]);
  TRQ = mxGetPr(plhs[0]);
  DQB = mxGetPr(plhs[1]);

  trace_schur3qq(N,R,G,B,TRQ,DQB);
  return;
}
